@@ -1,22 +1,21 @@
 #include <ctype.h>
 #include <stdio.h>
 #include <string.h>
+
+#include "blkdev.h"
+#include "pathnames.h"
 #include "common.h"
+#include "c.h"
 
 /*
  * return partition name - uses static storage unless buf is supplied
  */
-static char *
-partnamebf(char *dev, int pno, int lth, int bufsiz, char *bufp) {
-	static char buffer[80];
+char *
+partname(char *dev, int pno, int lth) {
+	static char bufp[PATH_MAX];
 	char *p;
 	int w, wp;
 
-	if (!bufp) {
-		bufp = buffer;
-		bufsiz = sizeof(buffer);
-	}
-
 	w = strlen(dev);
 	p = "";
 
@@ -30,17 +29,21 @@ partnamebf(char *dev, int pno, int lth, int bufsiz, char *bufp) {
 		p = "part";
 	}
 
+	/* udev names partitions by appending -partN
+	   e.g. ata-SAMSUNG_SV8004H_0357J1FT712448-part1 */
+	if ((strncmp(dev, _PATH_DEV_BYID, strlen(_PATH_DEV_BYID)) == 0) ||
+	     strncmp(dev, _PATH_DEV_BYPATH, strlen(_PATH_DEV_BYPATH)) == 0) {
+	       p = "-part";
+	}
+
 	wp = strlen(p);
-		
+
 	if (lth) {
-		sprintf(bufp, "%*.*s%s%-2u", lth-wp-2, w, dev, p, pno);
+		snprintf(bufp, sizeof(bufp), "%*.*s%s%-2u",
+			 lth-wp-2, w, dev, p, pno);
 	} else {
-		sprintf(bufp, "%.*s%s%-2u", w, dev, p, pno);
+		snprintf(bufp, sizeof(bufp), "%.*s%s%-2u", w, dev, p, pno);
 	}
 	return bufp;
 }
 
-char *
-partname(char *dev, int pno, int lth) {
-	return partnamebf(dev, pno, lth, 0, NULL);
-}
