@@ -11,7 +11,7 @@
  * @short_description: container for entries from fstab, mtab or mountinfo
  *
  * Note that mnt_table_find_* functions are mount(8) compatible. These functions
- * try to found an entry in more iterations where the first attempt is always
+ * try to find an entry in more iterations where the first attempt is always
  * based on comparison with unmodified (non-canonicalized or un-evaluated)
  * paths or tags. For example fstab with two entries:
  * <informalexample>
@@ -44,6 +44,7 @@
 #include <blkid.h>
 
 #include "mountP.h"
+#include "strutils.h"
 
 /**
  * mnt_new_table:
@@ -285,6 +286,11 @@ int mnt_table_next_child_fs(struct libmnt_table *tb, struct libmnt_iter *itr,
 
 		id = mnt_fs_get_id(fs);
 
+		/* avoid infinite loop. This only happens in rare cases
+		 * such as in early userspace when the rootfs is its own parent */
+		if (id == parent_id)
+			continue;
+
 		if ((!lastchld_id || id > lastchld_id) &&
 		    (!*chld || id < chld_id)) {
 			*chld = fs;
@@ -437,7 +443,7 @@ struct libmnt_fs *mnt_table_find_target(struct libmnt_table *tb, const char *pat
 	/* native @target */
 	mnt_reset_iter(&itr, direction);
 	while(mnt_table_next_fs(tb, &itr, &fs) == 0) {
-		if (fs->target && strcmp(fs->target, path) == 0)
+		if (mnt_fs_streq_target(fs, path))
 			return fs;
 	}
 	if (!tb->cache || !(cn = mnt_resolve_path(path, tb->cache)))
@@ -446,7 +452,7 @@ struct libmnt_fs *mnt_table_find_target(struct libmnt_table *tb, const char *pat
 	/* canonicalized paths in struct libmnt_table */
 	mnt_reset_iter(&itr, direction);
 	while(mnt_table_next_fs(tb, &itr, &fs) == 0) {
-		if (fs->target && strcmp(fs->target, cn) == 0)
+		if (mnt_fs_streq_target(fs, cn))
 			return fs;
 	}
 
@@ -455,12 +461,13 @@ struct libmnt_fs *mnt_table_find_target(struct libmnt_table *tb, const char *pat
 	while(mnt_table_next_fs(tb, &itr, &fs) == 0) {
 		char *p;
 
-		if (!fs->target || !(fs->flags & MNT_FS_SWAP) ||
+		if (!fs->target || mnt_fs_is_swaparea(fs) ||
 		    (*fs->target == '/' && *(fs->target + 1) == '\0'))
 		       continue;
 
 		p = mnt_resolve_path(fs->target, tb->cache);
-		if (strcmp(cn, p) == 0)
+		/* both canonicalized, strcmp() is fine here */
+		if (p && strcmp(cn, p) == 0)
 			return fs;
 	}
 	return NULL;
@@ -479,9 +486,8 @@ struct libmnt_fs *mnt_table_find_target(struct libmnt_table *tb, const char *pat
  * The 2nd, 3rd and 4th iterations are not performed when @tb cache is not
  * set (see mnt_table_set_cache()).
  *
- * Note that valid source path is NULL; the libmount uses NULL instead of
- * "none".  The "none" is used in /proc/{mounts,self/mountninfo} for pseudo
- * filesystems.
+ * Note that NULL is a valid source path; it will be replaced with "none". The
+ * "none" is used in /proc/{mounts,self/mountinfo} for pseudo filesystems.
  *
  * Returns: a tab entry or NULL.
  */
@@ -500,16 +506,10 @@ struct libmnt_fs *mnt_table_find_srcpath(struct libmnt_table *tb, const char *pa
 	/* native paths */
 	mnt_reset_iter(&itr, direction);
 	while(mnt_table_next_fs(tb, &itr, &fs) == 0) {
-		const char *src = mnt_fs_get_source(fs);
-
-		p = mnt_fs_get_srcpath(fs);
-
-		if (path == NULL && src == NULL)
-			return fs;			/* source is "none" */
-		if (p && strcmp(p, path) == 0)
+		if (mnt_fs_streq_srcpath(fs, path))
 			return fs;
-		if (!p && src)
-			ntags++;			/* mnt_fs_get_srcpath() returs nothing, it's TAG */
+		if (mnt_fs_get_tag(fs, NULL, NULL) == 0)
+			ntags++;
 	}
 
 	if (!path || !tb->cache || !(cn = mnt_resolve_path(path, tb->cache)))
@@ -519,8 +519,7 @@ struct libmnt_fs *mnt_table_find_srcpath(struct libmnt_table *tb, const char *pa
 	if (ntags < mnt_table_get_nents(tb)) {
 		mnt_reset_iter(&itr, direction);
 		while(mnt_table_next_fs(tb, &itr, &fs) == 0) {
-			p = mnt_fs_get_srcpath(fs);
-			if (p && strcmp(p, cn) == 0)
+			if (mnt_fs_streq_srcpath(fs, cn))
 				return fs;
 		}
 	}
@@ -551,7 +550,9 @@ struct libmnt_fs *mnt_table_find_srcpath(struct libmnt_table *tb, const char *pa
 				 if (mnt_fs_get_tag(fs, &t, &v))
 					 continue;
 				 x = mnt_resolve_tag(t, v, tb->cache);
-				 if (x && !strcmp(x, cn))
+
+				 /* both canonicalized, strcmp() is fine here */
+				 if (x && strcmp(x, cn) == 0)
 					 return fs;
 			 }
 		}
@@ -561,12 +562,14 @@ struct libmnt_fs *mnt_table_find_srcpath(struct libmnt_table *tb, const char *pa
 	if (ntags <= mnt_table_get_nents(tb)) {
 		mnt_reset_iter(&itr, direction);
 		while(mnt_table_next_fs(tb, &itr, &fs) == 0) {
-			if (fs->flags & (MNT_FS_NET | MNT_FS_PSEUDO))
+			if (mnt_fs_is_netfs(fs) || mnt_fs_is_pseudofs(fs))
 				continue;
 			p = mnt_fs_get_srcpath(fs);
 			if (p)
 				p = mnt_resolve_path(p, tb->cache);
-			if (p && strcmp(cn, p) == 0)
+
+			/* both canonicalized, strcmp() is fine here */
+			if (p && strcmp(p, cn) == 0)
 				return fs;
 		}
 	}
@@ -701,16 +704,18 @@ struct libmnt_fs *mnt_table_find_pair(struct libmnt_table *tb, const char *sourc
 }
 
 /*
- * @tb: /proc/self/mountinfo
- * @fs: filesystem
- * @mountflags: MS_BIND or 0
- * @fsroot: fs-root that will be probably used in the mountinfo file
+ * tb: /proc/self/mountinfo
+ * fs: filesystem
+ * mountflags: MS_BIND or 0
+ * fsroot: fs-root that will be probably used in the mountinfo file
  *          for @fs after mount(2)
  *
  * For btrfs subvolumes this function returns NULL, but @fsroot properly set.
  *
  * Returns: entry from @tb that will be used as a source for @fs if the @fs is
  *          bindmount.
+ *
+ * Don't export to library API!
  */
 struct libmnt_fs *mnt_table_get_fs_root(struct libmnt_table *tb,
 					struct libmnt_fs *fs,
@@ -730,19 +735,23 @@ struct libmnt_fs *mnt_table_get_fs_root(struct libmnt_table *tb,
 
 	if (tb && (mountflags & MS_BIND)) {
 		const char *src, *src_root;
+		char *xsrc = NULL;
 
 		DBG(TAB, mnt_debug("fs-root for bind"));
 
-		src = mnt_resolve_spec(mnt_fs_get_source(fs), tb->cache);
-		if (!src)
-			goto err;
+		src = xsrc = mnt_resolve_spec(mnt_fs_get_source(fs), tb->cache);
+		if (src)
+			mnt = mnt_get_mountpoint(src);
+		if (mnt)
+			root = mnt_get_fs_root(src, mnt);
 
-		mnt = mnt_get_mountpoint(src);
+		if (xsrc && !tb->cache) {
+			free(xsrc);
+			src = NULL;
+		}
 		if (!mnt)
 			goto err;
 
-		root = mnt_get_fs_root(src, mnt);
-
 		src_fs = mnt_table_find_target(tb, mnt, MNT_ITER_BACKWARD);
 		if (!src_fs)  {
 			DBG(TAB, mnt_debug("not found '%s' in mountinfo -- using default", mnt));
@@ -809,6 +818,20 @@ err:
 	return NULL;
 }
 
+static int is_mountinfo(struct libmnt_table *tb)
+{
+	struct libmnt_fs *fs;
+
+	if (!tb)
+		return 0;
+
+	fs = list_first_entry(&tb->ents, struct libmnt_fs, ents);
+	if (fs && mnt_fs_is_kernel(fs) && mnt_fs_get_root(fs))
+		return 1;
+
+	return 0;
+}
+
 /**
  * mnt_table_is_mounted:
  * @tb: /proc/self/mountinfo file
@@ -824,46 +847,70 @@ err:
 int mnt_table_is_fs_mounted(struct libmnt_table *tb, struct libmnt_fs *fstab_fs)
 {
 	char *root = NULL;
-	struct libmnt_fs *src_fs;
-	const char *src, *tgt;
-	int flags = 0, rc = 0;
+	const char *src = NULL;
+	char *xsrc = NULL, *tgt;
+	int rc = 0;
 
 	assert(tb);
 	assert(fstab_fs);
 
-	if (fstab_fs->flags & MNT_FS_SWAP)
+	if (mnt_fs_is_swaparea(fstab_fs))
 		return 0;
 
-	if (mnt_fs_get_option(fstab_fs, "bind", NULL, NULL) == 0)
-		flags = MS_BIND;
+	if (is_mountinfo(tb)) {
+		/* @tb is mountinfo, so we can try to use fs-roots */
+		struct libmnt_fs *fs;
+		int flags = 0;
 
-	src_fs = mnt_table_get_fs_root(tb, fstab_fs, flags, &root);
-	if (src_fs)
-		src = mnt_fs_get_srcpath(src_fs);
+		if (mnt_fs_get_option(fstab_fs, "bind", NULL, NULL) == 0)
+			flags = MS_BIND;
+
+		fs = mnt_table_get_fs_root(tb, fstab_fs, flags, &root);
+		if (fs)
+			src = mnt_fs_get_srcpath(fs);
+	}
+
+	if (src)
+		src = xsrc = mnt_resolve_spec(src, tb->cache);
+	else if (mnt_fs_is_pseudofs(fstab_fs))
+		src = mnt_fs_get_source(fstab_fs);
 	else
-		src = mnt_resolve_spec(mnt_fs_get_source(fstab_fs), tb->cache);
+		src = xsrc = mnt_resolve_spec(mnt_fs_get_source(fstab_fs),
+					      tb->cache);
 
-	tgt = mnt_fs_get_target(fstab_fs);
+	tgt = mnt_resolve_path(mnt_fs_get_target(fstab_fs), tb->cache);
 
-	if (tgt || src || root) {
+	if (tgt && src) {
 		struct libmnt_iter itr;
 		struct libmnt_fs *fs;
 
 		mnt_reset_iter(&itr, MNT_ITER_FORWARD);
 
 		while(mnt_table_next_fs(tb, &itr, &fs) == 0) {
-			const char *s = mnt_fs_get_srcpath(fs),
-				   *t = mnt_fs_get_target(fs),
-				   *r = mnt_fs_get_root(fs);
 
-			if (s && t && r && !strcmp(t, tgt) &&
-			    !strcmp(s, src) && !strcmp(r, root))
+			if (root) {
+				/* mountinfo: compare root, source and target */
+				const char *r = mnt_fs_get_root(fs);
+
+				if (r && strcmp(r, root) == 0 &&
+				    mnt_fs_streq_srcpath(fs, src) &&
+				    mnt_fs_streq_target(fs, tgt))
+					break;
+			}
+			/* mtab: compare source and target */
+			else if (mnt_fs_streq_srcpath(fs, src) &&
+				 mnt_fs_streq_target(fs, tgt))
 				break;
 		}
 		if (fs)
 			rc = 1;		/* success */
 	}
 
+	if (xsrc && !tb->cache)
+		free(xsrc);
+	if (!tb->cache)
+		free(tgt);
+
 	free(root);
 	return rc;
 }
